#!/usr/bin/perl

# Evaluate splice site substitution patterns along a phylogenetic tree
# (proof of concept implementation)

# Author:  Dominic Rose
# Contact: dominic@bioinf.uni-leipzig.de
#          (University of Leipzig)

use strict;
use Getopt::Long;
use Math::Round;
use Bio::TreeIO;
use Bio::Tree::Draw::Cladogram;

my $usage = << "JUS";

  Evaluate splice site substitution patterns along a phylogenetic tree

  USAGE: perl $0 -s FILE -m FILE -d FILE [-c INT] [--v|verbose] [-tmp FILE]

  Author:   Dominic Rose, University of Leipzig
  Contact:  dominic\@bioinf.uni-leipzig.de

  Citation: Computational discovery of human coding and non-coding
            transcripts with conserved splice sites.
            Rose D, Hiller M, Schutt K, Hackermueller J, Backofen R and Stadler, PF.
            Bioinformatics 2011 May 26.

  OPTIONS: 
	-s FILE	     FASTA file (input) containing your splice site candidate
                     sequences.
                     We allow to process multiple fasta alignments seperated 
                     by an empty line in the same file. 
                     So please ensure that the last line of your input is an 
                     empty line. 
                     Otherwise the script misses your last entry.
		     [REQUIRED]

    --- OPTIONAL ---

        -c INT       Print details of the specified column (implies --verbose).
                     Either within [-5,3] for donors or [-19,1] for acceptors.
                     These intervals are 0-based. Thus, 0 is the G of GT/AG.
                     [OPTIONAL]

        -e FILE      Print the resulting tree as EPS into the specified file.
                     Just for visualization.
                     [OPTIONAL]

        -tmp DIR     Directory for temporary files
                     [OPTIONAL, default: '/var/tmp']

        -l FILE      File containing the precomputed log-odds scores.
                     Delivered by us.
                     Format: A 3 1 -19 AC -0.4211
                     [OPTIONAL, default: './model/tree-based.log-odds.dat']

	-m FILE      .mod file from UCSC, contains tree etc.
                     Delivered by us.
		     [OPTIONAL, default: './model/vertebrate.mod']

        -d FILE      .dnd file, the UCSC tree in newick format
                     Delivered by us.
		     [OPTIONAL, default: './model/vertebrate.dnd']

        -p           Path of your prequel binary.
                     Please adjust this according to your environment.
                     Prequel is part of the phast package available at
                     http://compgen.bscb.cornell.edu/phast/
                     [OPTIONAL, default: 'prequel32']

        --v|verbose  Verbose mode
                     [OPTIONAL]

  PURPOSE:
	Score substitutions along a phylogenetic tree 
        using pre-computed log-odd substitution scores.

  DRAWBACK:
        The model underlying this script was built on the 
        multiz 44-way alignment available at the UCSC Genome Browser.
        See ftp://hgdownload.cse.ucsc.edu/goldenPath/hg18/multiz44way/

        It relies on the UCSC 44-way species and the corresponding tree.
        Therefore, this script can only handle alignments 
        that contain these species.

        Please ensure that you provide correct species names according to 
        the UCSC conventions, e.g. hg18.chr1 for human chromosome 1.

  INPUT FORMAT:
        We currently only support multiple FASTA files, separated by newlines.
        Please note that we require an empty line as last line 
        for each FASTA file, since we recognize alignment blocks via this 
        empty newline.

  EXAMPLE DATA:
        You find some exemple fasta files in the data subdirectory.

  RESULTS:
	At STDOUT
JUS

# options
my $opt_s = '';
my $opt_m = './model/vertebrate.mod';
my $opt_d = './model/vertebrate.dnd';
my $opt_c = 0;
my $opt_l = './model/tree-based.log-odds.dat';
my $opt_e = '';
my $opt_tmp = '/var/tmp';
my $opt_p = 'prequel32';
my $opt_v;

GetOptions(
  "s=s"       => \$opt_s,
  "m:s"       => \$opt_m,
  "d:s"       => \$opt_d,
  "c:i"       => \$opt_c,
  "l:s"       => \$opt_l,
  "v|verbose" => \$opt_v,
  "e:s"       => \$opt_e,
  "tmp:s"     => \$opt_tmp,
  "p:s"       => \$opt_p,
  "<>"        => \&usage
);

if ( !$opt_s || !$opt_m || !$opt_d || !$opt_l || !$opt_p) {
  usage();
}

my %species; # valid species names
$species{hg18}=1;
$species{panTro2}=1;
$species{gorGor1}=1;
$species{ponAbe2}=1;
$species{rheMac2}=1;
$species{calJac1}=1;
$species{tarSyr1}=1;
$species{micMur1}=1;
$species{otoGar1}=1;
$species{tupBel1}=1;
$species{mm9}=1;
$species{rn4}=1;
$species{dipOrd1}=1;
$species{cavPor3}=1;
$species{speTri1}=1;
$species{oryCun1}=1;
$species{ochPri2}=1;
$species{vicPac1}=1;
$species{turTru1}=1;
$species{bosTau4}=1;
$species{equCab2}=1;
$species{felCat3}=1;
$species{canFam2}=1;
$species{myoLuc1}=1;
$species{pteVam1}=1;
$species{eriEur1}=1;
$species{sorAra1}=1;
$species{loxAfr2}=1;
$species{proCap1}=1;
$species{echTel1}=1;
$species{dasNov2}=1;
$species{choHof1}=1;
$species{monDom4}=1;
$species{ornAna1}=1;
$species{galGal3}=1;
$species{taeGut1}=1;
$species{anoCar1}=1;
$species{xenTro2}=1;
$species{tetNig1}=1;
$species{fr2}=1;
$species{gasAcu1}=1;
$species{oryLat2}=1;
$species{danRer5}=1;
$species{petMar1}=1;

# quick n dirty check if prequel is installed
my $tmp = `$opt_p -h`;
die "ERROR: Please install prequel first or provide the correct path to your prequel binary as command line parameter! " if($tmp !~ /PREQUEL/);

if ($opt_c) {
  $opt_v = 1;
  if ( $opt_c <= -19 || $opt_c > 5 ) {
    die "ERROR: check -c parameter, $opt_c is not allowed!";
  }
}

my $t;    # tree pointer (full tree)

my $this_t;         # small tree of current alignment
my %map;            # iid -> 'id' = names (of the original UCSC tree)
                    # iid -> 'seq' = sequence
my %thisSpecies;    # names of the current species
my %thisNodes;      # names -> id, just a helper
my @fa;             # header, seq, header, seq, ...

my ( $tmpfile1, $tmpfile2, $tmpfile3 ) = ( "", "", "" );
my $header = "";
my $what;           # D or A
my %subst;          # hash that stores the substitution info
my %valid = (
  "AT" => 1,
  "AG" => 1,
  "AC" => 1,
  "TA" => 1,
  "TG" => 1,
  "TC" => 1,
  "GA" => 1,
  "GT" => 1,
  "GC" => 1,
  "CA" => 1,
  "CG" => 1,
  "CT" => 1
);
my %precomputed;    # hash of log-odds

my $do_length   = 9;
my $acc_length  = 21;
my $do_shift    = 3;
my $acc_shift   = 19;
my $relPosShift = 0;
my $l           = 0;    # length of analysed sequence (length of D or ACC).
my $qt_round    = 0;
my $total_score = 0;    # final result
my $whole_line  = '';
my $qt_subst    = 0;

########## go ##############
readPrecomputedLogOdds();
setupTreeHash();        # fill $t
parseFasta();

printTree() if ($opt_e);

unlink($tmpfile1);
unlink($tmpfile2);
unlink($tmpfile3);

########## methods #########

sub usage {
  print STDERR $usage;
  exit;
}

sub printTree {
  my $obj1 = Bio::Tree::Draw::Cladogram->new(
    -bootstrap => 0,
    -tree      => $this_t,
    -compact   => 1
  );
  $obj1->print( -file => $opt_e );
}

# read the log-odds
sub readPrecomputedLogOdds {
  open IN, "< $opt_l" or die "ERROR: Could not open file '$opt_l'!";
  while (<IN>) {
    my @e = split / /, $_;
    $precomputed{ $e[0] }{ $e[1] }{ $e[2] }{ $e[3] }{ $e[4] } = $e[5];
  }
  close IN;
}

sub printScore {
  $total_score = nearest( .01, $total_score );

  print "Final tree-based log-odds score: " if ($opt_v);
  print $total_score, "\n";
}

sub parseSmallTree {
  my $root = $this_t->get_root_node;

  my $leaves = 0;
  my $no     = 1;    # root
  if ($opt_v) {
    print $root->internal_id, " ", $root->id, " (root)\n";
    foreach my $node ( $root->get_all_Descendents ) {
      $no++;
      print $node->internal_id, " ", $node->id;
      if ( $node->is_Leaf ) {
        $leaves++;
        print " (leave) ";
      }
      print "\n";
    }
  }

  # now compare sequences of each node with its child seqs.
  my @todo = ();
  push @todo, $root;
  push @todo, $root->get_all_Descendents;
  my $qtEdges = 0;
  print "--- analyse substitutions ---\n" if ($opt_v);
  foreach my $node (@todo) {
    foreach my $k ( $node->each_Descendent() ) {    # kids
      $qtEdges++;
      print $map{ $node->internal_id }{'seq'} . " "
        . $map{ $node->internal_id }{'id'} . " "
        . $node->internal_id . "\n"
        . $map{ $k->internal_id }{'seq'} . " "
        . $map{ $k->internal_id }{'id'} . " "
        . $k->internal_id . "\n"
        if ($opt_v);

      # check for substitutions
      my @chars_parent = split //, $map{ $node->internal_id }{'seq'};
      my @chars_child  = split //, $map{ $k->internal_id }{'seq'};

      if ( scalar @chars_parent != scalar @chars_child ) {
        die "ERROR: Your sequences do not have the same length!\n";
      }

      my $edgeScore = 0;
      for ( my $i = 0 ; $i < $l ; $i++ ) {
        my $di = $chars_parent[$i] . $chars_child[$i];

        if ( exists $valid{$di} ) {
          my $col = ( $i - $relPosShift );

          if ( exists $precomputed{$what}{ $node->internal_id }{ $k->internal_id }{$col}{$di} ) {
            $qt_subst++;
            $total_score += $precomputed{$what}{ $node->internal_id }{ $k->internal_id }{$col}{$di};
          }

          if ($opt_v) {
            if ( exists $precomputed{$what}{ $node->internal_id }{ $k->internal_id }{$col}{$di} ) {
              print $what, " ", $node->internal_id, " ", $k->internal_id, " ", $col, " ", $di, " ";
              print $precomputed{$what}{ $node->internal_id }{ $k->internal_id }{$col}{$di};
              $edgeScore += $precomputed{$what}{ $node->internal_id }{ $k->internal_id }{$col}{$di};
            }
          }
        }
      }
      if ($opt_v) {
        print "edgeScore = ", nearest(0.01, $edgeScore), "\n\n";
      }
    }
  }

  print "Resulting tree: #leaves=$leaves #nodes=$no #edges=$qtEdges #substitutions=$qt_subst\n"
    if ($opt_v);
}

# sort by value
sub hashValueDescendingNum {
  $thisNodes{$b} <=> $thisNodes{$a};
}

# reduce the full UCSC tree to the tree given by the current alignment
# but let's do it on a copy
sub shrinkTree {
  $this_t = $t->clone();

  # delete leaves
  my $go   = 1;
  my $root = $this_t->get_root_node;

  while ($go) {
    foreach my $node ( $root->get_all_Descendents() ) {
      $go = 0;
      if ( $node->is_Leaf && ( !exists $thisNodes{ $node->id } ) ) {
        print "ignoring ", $node->id, "\n" if ($opt_v);
        $this_t->remove_Node( $node->id );
        $go = 1;
        $this_t->contract_linear_paths();
      }
    }
  }

  my @iids = sort { $b <=> $a } values %thisNodes;

  # reroot
  foreach my $iid (@iids) {
    if ( my $node = $this_t->find_node( -internal_id => $iid ) ) {
      if ( $node->internal_id != $root->internal_id ) {
        $this_t->reroot($node);
        print "new root: ", $node->internal_id, " ", $node->id, "\n" if ($opt_v);
      }
      $this_t->contract_linear_paths();
      last;    # only once
    }
  }

  # now remove all leaves that we have missed before rerooting
  $go = 1;
  while ($go) {
    my $root = $this_t->get_root_node;
    foreach my $node ( $root->get_all_Descendents() ) {
      $go = 0;
      if ( $node->is_Leaf && ( !exists $thisNodes{ $node->id } ) ) {
        print "ignoring ", $node->id, "\n" if ($opt_v);
        $this_t->remove_Node( $node->id );
        $go = 1;
        $this_t->contract_linear_paths();
      }
    }
  }

  $this_t->contract_linear_paths();    # remove the inner unnecessary nodes
}

sub runPrequel {
  if ( scalar @fa <= 1 ) {
    die "ERROR: Empty alignment file?";
  }

  my $tmp = $opt_s;
  $tmp =~ s/\//\_/g;
  $tmpfile2 = $opt_tmp . "/" . "tmp." . $tmp . ".fa";
  $tmpfile3 = $opt_tmp . "/" . "tmp." . $tmp . ".aln";
  open OUT, " > $tmpfile2 "
    or die " ERROR : Can not open temp file '$tmpfile2'";

  for ( my $i = 0 ; $i < $#fa ; $i++ ) {
    $fa[$i] = $1 if ( $fa[$i] =~ /^([^\.]+)\./ );    # header anpassen
    print OUT ">", $fa[$i], "\n", $fa[ ++$i ], "\n";
  }
  close OUT;

  my $cmd = "$opt_p -n $tmpfile2 $opt_m 2> /dev/null";
  my $ret = `$cmd` or die $?;

  my $lca;
  my @fa2 = split /\n/, $ret;
  print "----- reconstructed ancestral sequences : column(pos) seq iid species -----\n"
    if ($opt_v);
  foreach my $line (@fa2) {
    if ( $line =~ /^>(.+)$/ ) {
      my $tmp = $1;
      chomp $tmp;
      $header = $tmp;

      my @aa = split /-/, $header;
      my @nodes;
      my @n1 = $t->find_node( -id => $aa[0] );
      my @n2 = $t->find_node( -id => $aa[1] );
      die " ERROR : node not found for id $aa[0] . "
        if ( scalar(@n1) != 1 );
      die " ERROR : node not found for id $aa[1] . "
        if ( scalar(@n2) != 1 );
      push @nodes, $n1[0];
      push @nodes, $n2[0];

      if ( $lca = $t->get_lca(@nodes) ) {
        $thisNodes{$header} = $lca->internal_id;
      } else {
        die " ERROR : Can not find lca for ", $n1[0]->internal_id, " ", $n2[0]->internal_id;
      }
    } else {
      $map{ $lca->internal_id }{'seq'} = $line;
      $map{ $lca->internal_id }{'id'}  = $header;

      if ($opt_v) {
        my @char = split //, $line;
        print $char[ $relPosShift + $opt_c ], " ($opt_c) "
          if ($opt_c);
        print "$line $thisNodes{$header} $header\n";
      }
    }
  }
}

# read in fasta sequences of the SS file
sub parseFasta {
  my $node;
  open IN, "< $opt_s" or die "ERROR: Can not open infile '$opt_s'";
  while (<IN>) {
    if ( $_ =~ /^\#/ ) {
      next;
    } elsif ( $_ =~ /^>(.+)$/ ) {    # alignment header
      print "----- next alingnment, original sequences: column (pos) seq species -----\n"
        if ( $opt_v && scalar(@fa) == 0 );
      my $tmp = $1;
      chomp $tmp;
      $header = $tmp;
      push @fa, $header;

      $header =~ /^([^\.]+)\./;
      $node = $t->find_node( -id => $1 );
      $thisSpecies{$1} = 1;

die "Please only use species names of the 44 way alignment, the model was trained for these sequences! (invalid species: '$1')" if(!exists $species{$1});

      $thisNodes{$1}   = $node->internal_id;
    } elsif ( $_ =~ /^(\S+)$/ ) {    # sequences
      my $tmp = $1;
      chomp $tmp;
      my $seq = $1;
      $map{ $node->internal_id }{'seq'} = uc($seq);
      push @fa, $seq;
print Dumper @fa;

      if ( ( length $seq ) == 9 ) {    # [-5,3]
        $what        = 'D';
        $l           = $do_length;
        $relPosShift = $do_shift;
        if ( $opt_c != 0 ) {
          my @char = split //, $seq;
          print $char[ $relPosShift + $opt_c ], " (", $opt_c, ") "
            if ( $opt_c != 0 );
          print $seq, " ", $node->internal_id, " ", $header, "\n";
        }
      } elsif ( ( length $seq ) == 21 ) {    # [-19,1]
        $what        = 'A';
        $l           = $acc_length;
        $relPosShift = $acc_shift;
        if ($opt_v) {
          # -19,1 -> 0-20
          # 0 is 19 -6 is 19-6
          my @char = split //, $seq;
          print $char[ $relPosShift + $opt_c ], " (", $opt_c, ") "
            if ( $opt_c != 0 );
          print $seq, " ", $node->internal_id, " ", $header, "\n";
        }
      } else {
        die
          "ERROR: Wrong sequence length ($seq), ensure that you only provide [-5,3] for donors and [-19,1] for acceptors!";
      }
    }

    # process the read alignment and
    # reset everything for next alignment
    elsif ( $_ =~ /^\n/ ) {
      runPrequel();
      shrinkTree();
      parseSmallTree();
      printScore();

      #            $qt_round++;
      #            print STDERR "\r" . $qt_round if ( !$opt_v );

      %thisSpecies = ();
      %thisNodes   = ();
      @fa          = ();
      $total_score = 0;
      $qt_subst    = 0;

      %subst = ();
    }
  }
  close IN;
}

# for each node of the tree given by the mod file
# keep track of the subnodes
# hash structure: nodes -> subnodes
# trick: run prequel with artifical dataset to retrieve the (inner-)node-names.
sub setupTreeHash {
  my $treeString = "";
  open IN, " < $opt_m " or die " ERROR : Can not open file '$opt_m'";
  while (<IN>) {
    $treeString = $1 if ( $_ =~ /^TREE: (\S+)/ );
  }
  close IN;

  $treeString =~ s/\d+\.\d+//g;
  $treeString =~ s/://g;
  $treeString =~ s/[(|)|;]+//g;

  my @names = split /,/, $treeString;

  my $tmpString = "";
  my $i         = 0;
  foreach my $n (@names) {
    $i++;
    $tmpString .= "$n ^ A ";
    $tmpString .= " & " if ( $i <= $#names );
  }

  my $out = setupTreeWithPrequel($tmpString);

  my $treeio = new Bio::TreeIO( -format => 'newick', -file => $opt_d );

  if ( $t = $treeio->next_tree ) {
    my @e = split /\n/, $out;
    foreach my $e (@e) {
      if ( $e =~ /^>(\S+)/ ) {
        my $prequelName = $1;
        my @ee = split /-/, $prequelName;

        my @n0 = $t->find_node( -id => $ee[0] );
        my @n1 = $t->find_node( -id => $ee[1] );

        my @nodes;
        push @nodes, $n0[0];
        push @nodes, $n1[0];
        if ( my $lca = $t->get_lca(@nodes) ) {
          $lca->id($prequelName);
        } else {
          die " ERROR : Can not find lca for ", $n0[0]->id, " ", $n1[0];
        }
      }
    }
    my $rootnode = $t->get_root_node;
    $map{ $rootnode->internal_id }{'id'} = $rootnode->id;

    foreach my $node ( $rootnode->get_all_Descendents() ) {
      $map{ $node->internal_id }{'id'} = $node->id;
    }
  }

  if ($opt_v) {
    print "--- initial tree ---\n";
    foreach my $iid ( sort { $a <=> $b } keys %map ) {
      print $iid, " ", $map{$iid}{'id'}, "\n";
    }
  }
}

sub setupTreeWithPrequel {
  my ($alnString) = @_;

  my @e = split /&/, $alnString;

  my $tmp = $opt_s;
  $tmp =~ s/\//\_/g;
  $tmpfile1 = $opt_tmp . "/" . "tmp." . $tmp . ".fa";

  open OUT, " > $tmpfile1 "
    or die " ERROR : Can not open temp file '$tmpfile1'";
  foreach my $seq (@e) {
    my @s = split /\^/, $seq;
    $s[0] =~ /^([^\.]+)\.?/;
    $s[0] = $1;
    die " ERROR : Wasn't able to parse species and sequence !"
      if ( scalar(@s) != 2 );
    print OUT " > $s[0] \n $s[1] \n ";
  }

  close OUT;

  my $cmd = "$opt_p -n $tmpfile1 $opt_m 2> /dev/null";
  my $ret = `$cmd` or die $?;
  return $ret;
}
